"""
This configuration test module was taken from phys2bids.
Credit to the original author(s) and to the phys2bids community.
"""

import os
import ssl
from urllib.request import urlretrieve

import pytest


def fetch_file(osf_id, path, filename):
    """
    Fetches file located on OSF and downloads to `path`/`filename`1

    Parameters
    ----------
    osf_id : str
        Unique OSF ID for file to be downloaded. Will be inserted into relevant
        location in URL: https://osf.io/{osf_id}/download
    path : str
        Path to which `filename` should be downloaded. Ideally a temporary
        directory
    filename : str
        Name of file to be downloaded (does not necessarily have to match name
        of file on OSF)

    Returns
    -------
    full_path : str
        Full path to downloaded `filename`
    """
    # This restores the same behavior as before.
    # this three lines make tests dowloads work in windows
    if os.name == 'nt':
        orig_sslsocket_init = ssl.SSLSocket.__init__
        ssl.SSLSocket.__init__ = lambda *args, cert_reqs=ssl.CERT_NONE, **kwargs: orig_sslsocket_init(*args, cert_reqs=ssl.CERT_NONE, **kwargs)
        ssl._create_default_https_context = ssl._create_unverified_context
    url = 'https://osf.io/{}/download'.format(osf_id)
    full_path = os.path.join(path, filename)
    if not os.path.isfile(full_path):
        urlretrieve(url, full_path)
    return full_path


@pytest.fixture(scope='session')
def testdir(tmp_path_factory):
    """ Test path that will be used to download all files """
    return tmp_path_factory.getbasetemp()


@pytest.fixture
def atlas(testdir):
    return fetch_file('h6nj7', testdir,
                      'atlas.nii.gz')


@pytest.fixture
def mean_fc(testdir):
    return fetch_file('jrg8d', testdir,
                      'mean_fc_matlab.tsv')


@pytest.fixture
def sdi(testdir):
    return fetch_file('rs4dn', testdir,
                      'SDI_matlab.tsv')


@pytest.fixture
def sc_mtx(testdir):
    return fetch_file('vwh75', testdir,
                      'sc.mat')


@pytest.fixture
def timeseries(testdir):
    return fetch_file('ay8df', testdir,
                      'func.mat')



#!/usr/bin/env python3
"""
Operations for laplacian decomposition.

Attributes
----------
LGR
    Logger
"""

import logging

import numpy as np


LGR = logging.getLogger(__name__)


def symmetric_normalisation(mtx):
    """
    Compute symmetrically normalised Laplacian matrix.

    Parameters
    ----------
    mtx : numpy.ndarray
        A [structural] matrix

    Returns
    -------
    numpy.ndarray
        The symmetrically normalised version of mtx

    See Also
    --------
    https://en.wikipedia.org/wiki/Laplacian_matrix#Symmetrically_normalized_Laplacian_2
    """
    d = np.diag(mtx.sum(axis=-1) ** (-1/2))

    symm_norm = (d @ mtx @ d)

    return (np.eye(mtx.shape[0]) - symm_norm)


def decomposition(mtx):
    """
    Run a eigenvector decomposition on input.

    Parameters
    ----------
    mtx : numpy.ndarray
        The matrix to decompose

    Returns
    -------
    numpy.ndarray
        The eigenvalues resulting from the decomposition
    numpy.ndarray
        The eigenvectors resulting from the decomposition
    """
    eigenval, eigenvec = np.linalg.eig(mtx)

    idx = np.argsort(eigenval)
    eigenval = eigenval[idx]
    # #!# Check that eigenvec has the right index and not inverted 
    eigenvec = eigenvec[:, idx]

    return eigenval, eigenvec


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Visualisation utility.

Attributes
----------
FIGSIZE : tuple
    Figure size
LGR
    Logger
SET_DPI : int
    DPI of the figure
"""

import logging
import os

import numpy as np


LGR = logging.getLogger(__name__)
SET_DPI = 100
FIGSIZE = (18, 10)


def plot_connectivity(mtx, filename=None):
    """
    Create a connectivity matrix plot.

    If mtx has 3 dimensions, average first along the last axis.

    Parameters
    ----------
    mtx : numpy.ndarray
        A (square) array with connectivity information inside.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib is not installed.
    ValueError
        If mtx has more than 3 dimensions.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('matplotlib is required to plot connectivity matrices. '
                          'Please see install instructions.')

    mtx = mtx.squeeze()
    if mtx.ndim > 3:
        raise ValueError('Cannot plot connectivity matrices for matrix of '
                         'dimensions > 3.')
    elif mtx.ndim == 3:
        LGR.warning('Since matrix is 3D, averaging across last '
                    'dimension.')
        mtx = mtx.mean(axis=-1)

    if mtx.shape[0] != mtx.shape[1]:
        LGR.warning('Given matrix is not a square matrix!')

    LGR.info('Creating connectivity plot.')
    plt.figure(figsize=FIGSIZE)
    plt.imshow(mtx, cmap='RdBu')

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)

    return 0


def plot_grayplot(timeseries, filename=None):
    """
    Create a grayplot (a.k.a. carpet plot a.k.a. timeseries plot).

    If timeseries has 3 dimensions, average first along the last axis.

    Parameters
    ----------
    timeseries : numpy.ndarray
        An array representing a timeseries. Time has to be encoded in the
        second dimension.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib is not installed.
    ValueError
        If timeseries has more than 3 dimensions.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('matplotlib is required to plot grayplots. '
                          'Please see install instructions.')

    timeseries = timeseries.squeeze()
    if timeseries.ndim > 3:
        raise ValueError('Cannot plot grayplots for timeseries of '
                         'dimensions > 3.')
    elif timeseries.ndim == 3:
        LGR.warning('Since timeseries is 3D, averaging across last '
                    'dimension.')
        timeseries = timeseries.mean(axis=-1)

    LGR.info('Creating grayplot.')
    plt.figure(figsize=FIGSIZE)
    vmax = np.percentile(timeseries, 99)
    vmin = np.percentile(timeseries, 1)
    plt.imshow(timeseries, cmap='gray', vmin=vmin, vmax=vmax)

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
    return 0


def plot_nodes(ns, atlas, filename=None):
    """
    Create a marker plot in the MNI space.

    If ns has 2 dimensions, average first along last dimension.

    Parameters
    ----------
    ns : numpy.ndarray
        A 1- or 2- D array that contains the value of the nodes.
    atlas : str, os.PathLike, 3D Nifti1Image, or numpy.ndarray
        The 3d nifti image of an atlas, a string or path to its position,
        or a list of coordinates of the center of mass of parcels.
    filename : None, str, or os.PathLike, optional
        The path to save the plot on disk.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    ImportError
        If matplotlib and/or nilearn are not installed.
        If nibabel is not installed.
    nib.filebasedimages.ImageFileError
        If given path is not an atlas.
    ValueError
        If ns has more than 2 dimensions.
        If coordinates can't be extracted from atlas.
    """
    try:
        from nilearn.plotting import find_parcellation_cut_coords, plot_markers
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('nilearn and matplotlib are required to plot node images. '
                          'Please see install instructions.')

    # First check that ns is a valid source of data.
    ns = ns.squeeze()
    if ns.ndim > 2:
        raise ValueError('Cannot plot connectivity matrices for matrix of '
                         'dimensions > 2.')
    elif ns.ndim == 2:
        LGR.warning('Given matrix has 2 dimensions, averaging across last '
                    'dimension.')
        ns = ns.mean(axis=-1)

    # Then treat atlas
    try:
        coord = find_parcellation_cut_coords(atlas)
    except:
        if type(atlas) is np.ndarray:
            if atlas.ndim > 2 or atlas.shape[1] != 3:
                raise NotImplementedError('Only atlases in nifti format or '
                                          'list of coordinates are supported.')
            coord = atlas
        else:
            try:
                import nibabel as nib
                if os.path.isfile(atlas):
                    img = nib.load(atlas)
                else:
                    raise nib.filebasedimages.ImageFileError('Cannot find file '
                                                             f'{atlas}')

                coord = find_parcellation_cut_coords(img)
            except ImportError:
                raise ImportError('Nibabel is required to handle atlases I/O. '
                                  'Please see install instructions.')

    try:
        coord
    except NameError:
        raise ValueError('Could not obtain coordinates from given atlas.')

    if ns.shape[0] != coord.shape[0]:
        raise ValueError('Node array and coordinates array have different length.')

    LGR.info('Creating markerplot.')
    plt.figure(figsize=FIGSIZE)
    plot_markers(ns, coord)

    if filename is not None:
        plt.savefig(filename, dpi=SET_DPI)
    return 0


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
I/O and related utils.

Attributes
----------
EXT_1D : list
    List of supported TXT/1D file extensions, in lower case.
EXT_MAT : list
    List of supported matlab file extensions, in lower case.
EXT_NIFTI : list
    List of supported nifti file extensions, in lower case.
EXT_XLS : list
    List of supported XLS-like file extensions, in lower case.
LGR
    Logger
"""

import logging

from os import makedirs
from os.path import exists, join

import numpy as np

from nigsp.utils import change_var_type


EXT_1D = ['.txt', '.csv', '.tsv', '.1d', '.par', '.tsv.gz', '.csv.gz']
EXT_XLS = ['.xls']
EXT_MAT = ['.mat']
EXT_NIFTI = ['.nii', '.nii.gz']
EXT_ALL = EXT_1D + EXT_XLS + EXT_MAT + EXT_NIFTI

EXT_DICT = {'1D': EXT_1D, 'xls': EXT_XLS, 'mat': EXT_MAT, 'nifti': EXT_NIFTI}

LGR = logging.getLogger(__name__)


def check_ext(all_ext, fname, scan=False, remove=False):
    """
    Check which extension a file has, and possibly remove it.

    Parameters
    ----------
    all_ext : list
        All possible extensions to check within
    fname : str or os.PathLikeLike
        The filename to check
    scan : bool, optional
        Scan the given path to see if there is a file with that extension
        If True and no path declared, check if fname has a path, if not scan '.'
        If False, don't scan any folder
    remove : bool, optional
        Remove the extention from fname if it has one

    Returns
    -------
    obj_return : Uses a list to return variable amount of options.
        has_ext : boolean
            True if the extension is found, false otherwise
        fname : str or os.PathLike
            If `remove` is True, return (extensionless) fname
        ext : str
            If both `remove` and `has_ext` are True, returns also found extension
    """
    has_ext = False
    all_ext = change_var_type(all_ext, list, stop=False, silent=True)
    for ext in all_ext:
        if fname.lower().endswith(ext):
            has_ext = True
            LGR.debug(f'{fname} ends with extension {ext}')
            break

    if not has_ext and scan:
        for ext in all_ext:
            if exists(f'{fname}{ext}'):
                fname = f'{fname}{ext}'
                LGR.warning(f'Found {fname}{ext}, using it as input henceforth')
                has_ext = True
                break

    obj_return = [has_ext]

    if remove:
        if has_ext:
            obj_return += [fname[:-len(ext)], ext]  # case insensitive solution
        else:
            obj_return += [fname, '']
    else:
        obj_return += [fname]

    return obj_return[:]


def check_nifti_dim(fname, data, dim=4):
    """
    Check number of dimensions in nifti file.

    Parameters
    ----------
    fname : str
        The name of the file representing `data`
    data : numpy.ndarray
        The data which dimensionality needs to be checked
    dim : int, optional
        The amount of dimensions expected/desired in the data.

    Returns
    -------
    numpy.ndarray
        If `data.ndim` = `dim`, returns data.

    Raises
    ------
    ValueError
        If `data` has different dimensions than `dim`
    """
    data = data.squeeze()

    if data.ndim != dim:
        raise ValueError(f'A {dim}D nifti file is required, but {fname} is '
                         f'{data.ndim}D. Please check the input file.')

    return data


def check_mtx_dim(fname, data, shape=None):
    """
    Check dimensions of a matrix.

    Parameters
    ----------
    fname : str
        The name of the file representing `data`
    data : np.ndarray
        The data which dimensionality needs to be checked
    shape : None, 'square', or 'rectangle'}, str, optional
        Shape of matrix, if empty, skip shape check

    Returns
    -------
    np.ndarray
        If `data.ndim` = 2, returns data.
        If `data.ndim` = 1 and `shape` == 'rectangle', 
        Returns data with added empty axis.

    Raises
    ------
    NotImplementedError
        If `data` has more than 3 dimensions.
        If `shape` is not None but `data` is 3D.
    ValueError
        If `data` is empty
        If `shape` == 'square' and `data` dimensions have different lenghts.
    """
    data = data.squeeze()
    LGR.info('Checking data shape.')

    if data.shape[0] == 0:
        raise ValueError(f'{fname} is empty!')
    if data.ndim > 3:
        raise NotImplementedError('Only matrices up to 3D are supported, but '
                                  f'given matrix is {data.ndim}D.')
    if shape is not None:
        if data.ndim > 2:
            raise NotImplementedError('Cannot check shape of 3D matrix.')
        if data.ndim == 1 and shape == 'rectangle':
            data = data[..., np.newaxis]
            LGR.warning(f'Rectangular matrix required, but {fname} is a vector. '
                        'Adding empty dimension.')
        if shape == 'square' and data.shape[0] != data.shape[1]:
            raise ValueError(f'Square matrix required, but {fname} matrix has '
                             f'shape {data.shape}.')

    return data


def load_nifti_get_mask(fname, is_mask=False, ndim=4):
    """
    Load a nifti file and returns its data, its image, and a 3d mask.

    Parameters
    ----------
    fname : str
        The filename to read in
    is_mask : bool, optional
        If the file contains a mask.
        Default: False
    ndim : int
        The number of dimensions expected in the mask

    Returns
    -------
    data : numpy.ndarray
        Data from nifti file.
    mask : numpy.ndarray
        If `is_mask` is False, numpy.ndarray of one dimension less than data,
        in which any element that has at least a value different from zero
        in the last dimension of `data` is True.
        If `is_mask` is True, mask is a boolean representation of data.
    img : nib.img
        Image object from nibabel.

    """
    try:
        import nibabel as nib
    except ImportError:
        raise ImportError('nibabel is required to import nifti files. '
                          'Please see install instructions.')

    LGR.info(f'Loading {fname}.')
    img = nib.load(fname)
    data = img.get_fdata()
    data = check_nifti_dim(fname, data, dim=ndim)

    if is_mask:
        mask = (data != 0)
        LGR.info(f'{fname} loaded as mask.')
    else:
        mask = data.any(axis=-1).squeeze()
        LGR.info(f'Data loaded from {fname}.')

    return data, mask, img


def load_txt(fname, shape=None):
    """
    Read files in textual format.

    Parameters
    ----------
    fname : str or os.PathLike
        Path to the txt file
    shape : None, 'square', or 'rectangle', optional
        Shape of matrix, if empty, skip check

    Returns
    -------
    mtx : numpy.ndarray
        Data matrix

    See also
    --------
    check_mtx_dim
    """
    LGR.info(f'Loading {fname}.')

    _, _, ext = check_ext(EXT_1D, fname, scan=True, remove=True)

    if ext in ['.csv', '.csv.gz']:
        delimiter = ','
    elif ext in ['.tsv', '.tsv.gz']:
        delimiter = '\t'
    elif ext in ['.txt', '.1d', '.par']:
        delimiter = ' '
    else:
        delimiter = None

    mtx = np.genfromtxt(fname, delimiter=delimiter)

    mtx = check_mtx_dim(fname, mtx, shape)

    return mtx


def load_mat(fname, shape=None):
    """
    Read files in matlab format.

    Assumes the existence of a matrix/vector in the mat file, rendered as 
    a numpy.ndarray. If there is more than a marix, the one with the largest 
    size will be selected.

    Parameters
    ----------
    fname : str or os.PathLike
        Path to the mat file
    shape : None, 'square', or 'rectangle'}, str, optional
        Shape of matrix, if empty, skip check

    Returns
    -------
    mtx : numpy.ndarray
        Data matrix

    Notes
    -----
    Requires module pymatreader to work

    See also
    --------
    check_mtx_dim

    Raises
    ------
    EOFError
        If the mat file does not contain matrix or vectors.
    ImportError
        If pymatreader is not installed or can't be read.
    """
    try:
        from pymatreader import read_mat
    except ImportError:
        raise ImportError('pymatreader is required to import mat files. '
                          'Please see install instructions.')

    LGR.info(f'Loading {fname}.')
    data = read_mat(fname)

    data_keys = []
    for k in data.keys():
        # Check data key only if it's not hidden
        # (skip '__header__', '__version__', '__global__')
        if '__' not in k:
            LGR.info(f'Checking {fname} key {str(k)} content for data (float array/matrices in MATLAB).')
            if type(data[k]) is np.ndarray:
                data_keys.append(k)

    if len(data_keys) < 1:
        raise EOFError(f'{fname} does not seem to contain a numeric matrix.')
    elif len(data_keys) > 1:
        LGR.warning('Found multiple possible arrays to load. '
                    'Selecting the biggest (highest pythonic size).')

    key = data_keys[0]
    for k in data_keys[1:]:
        if data[k].size > data[key].size:
            key = k

    LGR.info(f'Selected data from MATLAB variable {key}')
    mtx = data[key]
    mtx = check_mtx_dim(fname, mtx, shape)

    return mtx


def load_xls(fname, shape=''):
    """
    Read files in xls format.

    Parameters
    ----------
    fname : str or os.PathLike
        Path to the mat file
    shape : None, 'square', or 'rectangle'}, str, optional
        Shape of matrix, if empty, skip check

    Notes
    -----
    Requires module _ to work

    See also
    --------
    check_mtx_dim

    No Longer Returned
    ------------------
    mtx : numpy.ndarray
        Data matrix

    Raises
    ------
    NotImplementedError
        Spreadheet loading is not implemented yet.
    """
    raise NotImplementedError('Spreadsheet loading is not implemented yet')


def export_nifti(data, img, fname):
    """
    Export a nifti file.

    Parameters
    ----------
    data : numpy.ndarray
        Data to be exported
    img : nib.img
        Nibabel image object
    fname : str or os.PathLike
        Name of the output file
    """
    try:
        import nibabel as nib
    except ImportError:
        raise ImportError('nibabel is required to export nifti files. '
                          'Please see install instructions.')

    for e in EXT_NIFTI:
        has_ext, fname, ext = check_ext(e, fname, remove=True)
        if has_ext:
            break

    if ext == '':
        ext = '.nii.gz'

    LGR.info(f'Exporting nifti data into {fname}{ext}.')
    out_img = nib.Nifti1Image(data, img.affine, img.header)
    out_img.to_filename(f'{fname}{ext}')

    return 0


def export_txt(data, fname, ext=None):
    """
    Export data into a text-like or mat file.

    Parameters
    ----------
    data : np.ndarray
        Data to be exported.
    fname : str or os.PathLike
        Name of the output file.
    ext : str or None, optional
        Selected extension for export.

    Returns
    -------
    0
        On a successful run
    """
    if ext.lower() in ['.csv', '.csv.gz', '', None]:
        delimiter = ','
    elif ext.lower() in ['.tsv', '.tsv.gz']:
        delimiter = '\t'
    elif ext.lower() in ['.txt', '.1d', '.par']:
        delimiter = ' '
    else:
        delimiter = None

    if data.ndim < 3:
        np.savetxt(f'{fname}{ext}', data, fmt='%.6f', delimiter=delimiter)
    elif data.ndim == 3:
        makedirs(fname, exist_ok=True)
        for i in range(data.shape[-1]):
            np.savetxt(join(fname, f'{i:03d}{ext}'), data[:, :, i], fmt='%.6f',
                       delimiter=delimiter)

    return 0


def export_mtx(data, fname, ext=None):
    """
    Export data into a text-like or mat file.

    Parameters
    ----------
    data : np.ndarray
        Data to be exported.
    fname : str or os.PathLike
        Name of the output file.
    ext : str or None, optional
        Selected extension for export.

    Notes
    -----
    Requires module scipy to export in .mat format.
    (Will require other modules to export in XLS-like format)

    Raises
    ------
    BrokenPipeError
        If somewhat an extension that is not supported passes all checks.
        (This should never happen)
    ImportError
        If scipy is not installed or cannot be found.
    NotImplementedError
        Spreadheet output is not implemented yet.

    Returns
    -------
    0
        On a successful run
    """
    if ext is None:
        # Check if extension was provided in fname.
        for e in EXT_ALL:
            has_ext, fname, ext = check_ext(e, fname, remove=True)
            if has_ext:
                break
    elif ext.lower() not in EXT_ALL:
        # Check if extension is supported.
        ext = None

    if ext in [None, '']:
        LGR.warning('Extension not specified, or specified extension not '
                    'supported. Forcing export in CSV format.')
        ext = '.csv'
    elif ext.lower() in EXT_NIFTI:
        LGR.warning('Found nifti extension, exporting data in .1D instead')
        ext = '.1D'

    LGR.info(f'Exporting data into {fname}{ext}.')
    if ext.lower() in EXT_MAT:
        try:
            import scipy
        except ImportError:
            raise ImportError('To export .mat files, scipy is required. '
                              'Please install it.')
        scipy.io.savemat(f'{fname}{ext}', {'data': data})
    elif ext.lower() in EXT_XLS:
        raise NotImplementedError('Spreadsheet output is not implemented yet')
    elif ext.lower() in EXT_1D:
        export_txt(data, fname, ext)
    else:
        raise BrokenPipeError(f'This should not have happened: {ext} was the '
                              'selected extension.')

    return 0


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Operations on graphs and graph derived properties.

Attributes
----------
LGR
    Logger
"""

import logging

from numpy import abs


LGR = logging.getLogger(__name__)


def zerocross(eigenvec):
    """
    Compute the amount of zero-crossing of an eigenvector matrix (for each eigenvector).

    Parameters
    ----------
    eigenvec : numpy.ndarray
        The eigenvectors from a decomposition.

    Returns
    -------
    numpy.ndarray
        A 1D array with the amount of zero-crossing for each eigenvector.
    """
    return (eigenvec[:-1, ...] * eigenvec[1:, ...] < 0).sum(axis=0)


def nodestrength(mtx, mean=False):
    """
    Compute the node strength of a graph.

    Parameters
    ----------
    mtx : numpy.ndarray
        A matrix depicting a graph.
    mean : bool, optional
        If True, return the average node strength along the last axis of mtx.

    Returns
    -------
    numpy.ndarray
        The node strength.
    """
    ns = abs(mtx).sum(axis=0)

    if mean:
        ns = ns.mean(axis=-1)

    return ns


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Toolbox objects.

Attributes
----------
LGR
    Logger
"""

import logging

from copy import deepcopy

from nigsp import operations


LGR = logging.getLogger(__name__)


class SCGraph():

    def __init__(self, mtx, timeseries, atlas=None, filename=None, img=None,
                 eigenval=None, eigenvec=None, energy=None, lapl_mtx=None,
                 index='median', evec_split=None, ts_split=None, surr=None,
                 surr_split=None, sdi=None, gsdi=None, fc=None, fc_split=None):
        """Initialise SCGraph (see class docstring)."""
        if mtx.shape[1] != mtx.shape[0]:
            raise ValueError('Graph matrix must be a square matrix, but '
                             f'given matrix has {mtx.shape}!')

        if mtx.shape[1] != timeseries.shape[0]:
            raise ValueError(f'Timeseries extracted from {timeseries.shape[0]} '
                             f'parcels, but graph has {mtx.shape[1]} nodes. '
                             'The number of parcels and nodes must be the same.')

        if timeseries.ndim > 3:
            raise ValueError('Timeseries of more than 3 dimensions are not supported '
                             f'yet, but given timeseries has {timeseries.ndim} '
                             'dimensions.')
        self.mtx = deepcopy(mtx)
        self.timeseries = deepcopy(timeseries)
        self.atlas = deepcopy(atlas)
        self.filename = deepcopy(filename)
        self.img = deepcopy(img)
        self.eigenval = deepcopy(eigenval)
        self.eigenvec = deepcopy(eigenvec)
        self.energy = deepcopy(energy)
        self.lapl_mtx = deepcopy(lapl_mtx)
        self.index = deepcopy(index)
        self.evec_split = deepcopy(evec_split)
        self.ts_split = deepcopy(ts_split)
        self.surr = deepcopy(surr)
        self.surr_split = deepcopy(surr_split)
        self.sdi = deepcopy(sdi)
        self.gsdi = deepcopy(gsdi)
        self.fc = deepcopy(fc)
        self.fc_split = deepcopy(fc_split)

    @property
    def nnodes(self):
        """Return number of nodes."""
        return self.mtx.shape[1]

    @property
    def ntimepoints(self):
        """Return number of timepoints."""
        return self.timeseries.shape[1]

    @property
    def zerocross(self):
        """Implement graph.zerocross as property."""
        return operations.zerocross(self.eigenvec)

    @property
    def split_keys(self):
        """Return the names of the splitted timeseries."""
        return list(self.ts_split.keys())

    def symmetric_normalisation(self):
        """Implement laplacian.symmetric_normalisation as class method."""
        self.lapl_mtx = operations.symmetric_normalisation(self.mtx)
        return self

    def decomposition(self):
        """Implement laplacian.decomposition as class method."""
        self.eigenval, self.eigenvec = operations.decomposition(self.lapl_mtx)
        return self

    def structural_decomposition(self):
        """Implement both laplacian operations."""
        return self.symmetric_normalisation().decomposition()

    def compute_graph_energy(self, mean=False):
        """Implement timeseries.graph_fourier_transform for energy as class method."""
        self.energy = operations.graph_fourier_transform(self.timeseries, self.eigenvec,
                                                         energy=True, mean=mean)
        return self

    def split_graph(self):
        """Implement timeseries.median_cutoff_frequency_idx as class method."""
        if self.index == 'median':
            self.index = operations.median_cutoff_frequency_idx(self.energy)
        else:
            raise ValueError(f'Unknown option {self.index} for frequency split index. '
                             'Declared index must be either an integer or \'median\'')
        self.evec_split, self.ts_split = operations.graph_filter(self.timeseries,
                                                                 self.eigenvec,
                                                                 self.index)

        return self

    def nodestrength(self, mean=False):
        """Implement graph.nodestrength as class method."""
        self.ns = operations.nodestrength(self.mtx, mean)
        return self

    def compute_sdi(self, mean=False, keys=None):
        """Implement sdi.sdi as class method."""
        self.sdi = operations.sdi(self.ts_split, mean, keys)
        return self

    def compute_gsdi(self, mean=False, keys=None):
        """Implement sdi.gsdi as class method."""
        self.gsdi = operations.gsdi(self.ts_split, mean, keys)
        return self

    def create_surrogates(self, sc_type='informed', n_surr=1000, seed=None):
        """Implement surrogates.sc_informed and sc_uninformed as class method."""
        sc_args = {'timeseries': self.timeseries, 'n_surr': n_surr}
        if seed is not None:
            sc_args['seed'] = seed

        if sc_type == 'informed':
            sc_args['eigenvec'] = self.eigenvec
            self.surr = operations.sc_informed(**sc_args)
        elif sc_type == 'uninformed':
            sc_args['lapl_mtx'] = self.lapl_mtx
            self.surr = operations.sc_uninformed(**sc_args)
        else:
            return ValueError(f'Unknown option {sc_type} for surrogate creation. '
                              'Declared type must be either \'informed\' or '
                              '\'uninformed\'')
        return self

    def test_significance(self, metric='sdi', method='Bernoulli', p=0.05,
                          return_masked=False, mean=False):
        """Implement surrogates.test_significance as class method."""
        _, self.surr_split = operations.graph_filter(self.surr,
                                                     self.eigenvec,
                                                     self.index)
        if metric == 'sdi':
            surr_sdi = operations.sdi(self.surr_split, mean, keys=None)
            metric_data = self.sdi
        elif metric == 'gsdi':
            surr_sdi = operations.gsdi(self.surr_split, mean, keys=None)
            metric_data = self.gsdi
        # #!# Check that this attribution works (should)
        metric_data = operations.test_significance(surr_sdi, metric_data,
                                                   method, p, return_masked,
                                                   mean)
        return self

    def normalise_ts(self):
        """Implement timeseries.normalise_ts for energy as class method."""
        self.timeseries = operations.normalise_ts(self.timeseries)
        return self

    def compute_fc(self, mean=False):
        """Implement timeseries.functional_connectivity as class method."""
        if self.timeseries is not None:
            LGR.info('Compute FC of original timeseries.')
            self.fc = operations.functional_connectivity(self.timeseries, mean)
        if self.ts_split is not None:
            self.fc_split = dict.fromkeys(self.ts_split.keys())
            LGR.info('Compute FC of split timeseries.')
            for k in self.ts_split.keys():
                LGR.info(f'Compute FC of {k} timeseries.')
                self.fc_split[k] = operations.functional_connectivity(self.ts_split[k],
                                                                      mean)
        return self


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
General utils.

Attributes
----------
LGR
    Logger
"""

import logging

from numpy import asarray, ndarray, prod, empty


LGR = logging.getLogger(__name__)


def change_var_type(var, dtype, varname='an input variable', stop=True,
                           silent=False):
    """
    Make sure `var` is of type `dtype`.

    Parameters
    ----------
    var : str, int, or float
        Variable to change type of
    dtype : type
        Type to change `var` to
    varname : str, optional
        The name of the variable
    stop : bool, optional
        If True, raises TypeError if `var` is not of `dtype`
    silent : bool, optional
        If True, don't return any message

    Returns
    -------
    int, float, str, list, or var
        The given `var` in the given `dtype`, or `var` if '' or None

    Raises
    ------
    NotImplementedError
        If dtype is not int, float, str, or list
    TypeError
        If variable var is not of type and stop is True
    """
    if type(var) is not dtype and stop:
        if varname != 'an input variable':
            varname = f'variable {varname}'
        raise TypeError(f'{varname} is not of type {dtype}')

    if dtype is int:
        tmpvar = int(var)
    elif dtype is float:
        tmpvar = float(var)
    elif dtype is str:
        tmpvar = str(var)
    elif dtype is ndarray:
        tmpvar = asarray(var)
    elif dtype is list:
        if type(var) is list:
            tmpvar = var
        else:
            tmpvar = [var]
    else:
        raise NotImplementedError(f'Type {dtype.__name__} not supported')

    if type(tmpvar) is not type(var):
        if varname != 'an input variable':
            varname = f'variable {varname}'
        msg = f'Changing type of {varname} from {type(var)} to {dtype}'

        if silent:
            LGR.debug(msg)
        else:
            LGR.warning(msg)

    return tmpvar


def prepare_ndim_iteration(data, idx):
    """
    Reshape data to have idx+1 dimensions.

    This should allow iterations over unknown numbers of dimensions above idx
    when needed.

    Parameters
    ----------
    data : np.ndarray
        The data to reiterate over
    idx : int
        The number of dimensions that should be fixed

    Returns
    -------
    np.ndarray, np.ndarray
        The reshaped data and an empty array like it.
    """
    if data.ndim > idx+1:
        new_shape = list(data.shape[:idx]) + [prod(data.shape[idx:])]
        temp_data = data.reshape(new_shape)
        temp_empty = empty(new_shape, dtype='float32')

        return temp_data, temp_empty
    else:
        return data, empty(data.shape, dtype='float32')


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Integration test."""
import shutil

from os.path import isfile, isdir, join

import numpy as np

from nigsp.workflow import _main


# ### Integration tests
def test_integration(timeseries, sc_mtx, atlas, mean_fc, sdi, testdir):
    """Integration test for FULL workflow."""
    testdir = join(testdir, 'testdir')
    mean_fc_mat = np.genfromtxt(mean_fc)
    sdi_mat = np.genfromtxt(sdi)

    # Run workflow
    _main(['-f', timeseries, '-s', sc_mtx, '-a', atlas, '-odir', testdir,
          '-ofile', 'testfile.tsv', '-sci', '-n', '4', '-seed', '42'])

    # Check that files were created
    assert isdir(testdir)
    assert isdir(join(testdir, 'testfile_fc'))
    assert isdir(join(testdir, 'testfile_fc_high'))
    assert isdir(join(testdir, 'testfile_fc_low'))
    assert isdir(join(testdir, 'logs'))
    assert isdir(join(testdir, 'testfile_timeseries_high'))
    assert isdir(join(testdir, 'testfile_timeseries_low'))
    assert isfile(join(testdir, 'testfile_fc', '000.tsv'))
    assert isfile(join(testdir, 'testfile_fc_high', '000.tsv'))
    assert isfile(join(testdir, 'testfile_fc_low', '000.tsv'))
    assert isfile(join(testdir, 'testfile_timeseries_high', '000.tsv'))
    assert isfile(join(testdir, 'testfile_timeseries_low', '000.tsv'))
    assert isfile(join(testdir, 'testfile_sc.png'))
    assert isfile(join(testdir, 'testfile_sdi.png'))
    assert isfile(join(testdir, 'testfile_fc.png'))
    assert isfile(join(testdir, 'testfile_grayplot.png'))
    assert isfile(join(testdir, 'testfile_fc_high.png'))
    assert isfile(join(testdir, 'testfile_grayplot_high.png'))
    assert isfile(join(testdir, 'testfile_sdi.tsv'))

    sdi_pyt = np.genfromtxt(join(testdir, 'testfile_sdi.tsv'))

    fc = np.empty((sdi_mat.shape[0], sdi_mat.shape[0], 10))
    for i in range(10):
        fc[..., i] = np.genfromtxt(join(testdir, 'testfile_fc', f'{i:03d}.tsv'))

    mean_fc_pyt = fc.mean(axis=-1)
    # Check that each cell in the result is comparable to matlab's.
    # There's a bunch of rounding due to np.round and numerical difference between matlab and python
    assert abs(mean_fc_pyt.round(6) - mean_fc_mat.round(6)).max().round(6) <= 0.000001
    assert abs(sdi_pyt.round(5) - sdi_mat.round(5)).max().round(5) <= 0.00001

    # Clean up!
    shutil.rmtree(testdir)


#!/usr/bin/env python3
"""Tests for utils."""

from numpy import ndarray
from numpy.random import rand
from pytest import mark, raises

from nigsp import utils


# ### Unit tests
@mark.parametrize('var, dtype', [
    (6, int),
    (4.2, float),
    ('hello', str),
    (rand(3), ndarray),
    ([1, 1, 2, 3, 5], list),
    ('6', int),
    (4, float),
    (42, str),
    ('hi', list),
    ([1, 1, 2, 3, 5], ndarray)
])
def test_change_var_type(var, dtype):
    """Test change_var_type."""
    # #!# Test logger!
    var_out = utils.change_var_type(var, dtype, stop=False)

    assert type(var_out) == dtype


# ### Break tests
@mark.parametrize('var, dtype', [
    ('6', int),
    (4, float),
    (42, str),
    ('hi', list),
    ([1, 1, 2, 3, 5], ndarray)
])
def test_break_change_var_type(var, dtype):
    """Break change_var_type."""
    with raises(TypeError) as errorinfo:
        utils.change_var_type(var, dtype)
    assert 'is not of type' in str(errorinfo.value)


def test_break_change_var_type_dtype():
    """Break change_var_type."""
    with raises(NotImplementedError) as errorinfo:
        utils.change_var_type(6, bool, stop=False)
    assert 'not supported' in str(errorinfo.value)



"""Import all operations."""

from .graph import zerocross, nodestrength
from .laplacian import symmetric_normalisation, decomposition
from .metrics import sdi, gsdi
from .nifti import (vol_to_mat, mat_to_vol, apply_mask, unmask, apply_atlas,
                    unfold_atlas)
from .surrogates import (random_sign, sc_informed, sc_uninformed,
                         test_significance)
from .timeseries import (normalise_ts, graph_fourier_transform,
                         median_cutoff_frequency_idx, graph_filter,
                         functional_connectivity)



#!/usr/bin/env python
import sys

from setuptools import setup
import versioneer

SETUP_REQUIRES = ['setuptools >= 30.3.0']
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

if __name__ == "__main__":
    setup(name='nigsp',
          setup_requires=SETUP_REQUIRES,
          version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass())



#!/usr/bin/env python3
"""Tests for operations.nifti."""
from numpy import prod, asarray, zeros, unique
from numpy.random import rand

from pytest import raises

from nigsp.operations import nifti


# ### Unit tests
def test_vol_to_mat():
    a = rand(3, 4, 5, 6)
    b = nifti.vol_to_mat(a)
    c = b.reshape(a.shape, order='F')

    assert b.ndim == 2
    assert b.shape == ((prod(a.shape[:-1])), a.shape[-1])
    assert (a == c).all()

    a = rand(3, 4, 5)
    b = nifti.vol_to_mat(a)
    c = b.reshape(a.shape, order='F')

    assert b.ndim == 1
    assert b.shape == prod(a.shape)
    assert (a == c).all()


def test_mat_to_vol():
    a = rand(3, 4, 5, 6)
    b = a.reshape((60, 6), order='F')
    c = nifti.mat_to_vol(b, asdata=a)

    assert c.ndim == 4
    assert c.shape == a.shape
    assert (a == c).all()

    c = nifti.mat_to_vol(b, shape=a.shape)

    assert c.ndim == 4
    assert c.shape == a.shape
    assert (a == c).all()

    c = nifti.mat_to_vol(b, shape=(6, 2, 5, 6), asdata=a)

    assert c.shape == a.shape


def test_apply_mask():
    m = asarray([0, 1, 0])
    a = rand(3, 3)

    b = nifti.apply_mask(a, m)
    assert (b == a[1, :]).all()


def test_unmask():
    m = asarray([0, 1, 0])
    a = rand(3)
    b = zeros([3, 3], dtype='float32')
    b[1, :] = a

    c = nifti.unmask(a, m, shape=b.shape)

    assert c.ndim == 2
    assert c.shape == b.shape
    assert (c == b).all()

    c = nifti.unmask(a, m, asdata=b)

    assert c.ndim == 2
    assert c.shape == b.shape
    assert (c == b).all()

    c = nifti.unmask(a, m, shape=(2, 2), asdata=b)

    assert c.shape == b.shape


def test_apply_atlas():
    m = asarray([0, 1, 1, 1, 1, 1])
    a = asarray([1, 2, 2, 3, 3, 0])
    d = rand(6, 10)
    c = zeros((3, 10), dtype='float32')
    cm = zeros((2, 10), dtype='float32')
    c[0, :] = d[0, :]
    c[1, :] = d[1:3, :].mean(axis=0)
    c[2, :] = d[3:5, :].mean(axis=0)
    cm[0, :] = d[1:3, :].mean(axis=0)
    cm[1, :] = d[3:5, :].mean(axis=0)

    r = nifti.apply_atlas(d, a)
    rm = nifti.apply_atlas(d, a, mask=m)

    assert r.ndim == 2
    assert r.shape == c.shape
    assert (r == c).all()
    assert rm.shape == cm.shape
    assert (rm == cm).all()


def test_unfold_atlas():
    m = asarray([0, 1, 1, 1, 1, 1])
    a = asarray([1, 2, 2, 3, 3, 0])
    c = rand(3, 10)
    cm = rand(2, 10)
    da = zeros((6, 10), dtype='float32')
    dm = zeros((6, 10), dtype='float32')
    label = unique(a)
    label = label[label > 0]
    for n, l in enumerate(label):
        da[a == l] = c[n, :]
    label = label[label > 1]
    for n, l in enumerate(label):
        dm[a == l] = cm[n, :]

    r = nifti.unfold_atlas(c, a)
    rm = nifti.unfold_atlas(cm, a, mask=m)

    assert r.ndim == 2
    assert r.shape == da.shape
    assert (r == da).all()
    assert rm.shape == dm.shape
    assert (rm == dm).all()


# ### Break tests
def test_break_mat_to_vol():
    with raises(ValueError) as errorinfo:
        nifti.mat_to_vol(rand(3))
    assert 'Both shape' in str(errorinfo.value)


def test_break_apply_mask():
    a = rand(3, 2, 5)
    m = rand(2, 2)
    with raises(ValueError) as errorinfo:
        nifti.apply_mask(a, m)
    assert f'shape {a.shape}' in str(errorinfo.value)
    assert f'shape {m.shape}' in str(errorinfo.value)


def test_break_unmask():
    a = rand(6, 5)
    m = rand(3)
    b = rand(2, 3, 5)

    with raises(ValueError) as errorinfo:
        nifti.unmask(a, m)
    assert 'shape and asdata' in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        nifti.unmask(a, m, shape=(2, 3))
    assert 'shape (2, 3)' in str(errorinfo.value)
    assert f'shape {m.shape}' in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        nifti.unmask(a, m, asdata=b)
    assert f'shape {b.shape}' in str(errorinfo.value)
    assert f'shape {m.shape}' in str(errorinfo.value)

    m = asarray([0, 0, 1, 0, 1, 1])
    with raises(ValueError) as errorinfo:
        nifti.unmask(a, m, shape=(6, 5))
    assert f'{m.sum()} entries' in str(errorinfo.value)


def test_break_apply_atlas():
    d = rand(2, 3, 4, 5)
    a = rand(3, 2, 4)
    b = rand(2, 3, 4)
    m = rand(3, 2, 4)
    with raises(NotImplementedError) as errorinfo:
        nifti.apply_atlas(d, d)
    assert f'Files with {d.ndim}' in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        nifti.apply_atlas(d, b, m)
    assert f'shape {d.shape}' in str(errorinfo.value)
    assert f'mask with shape {m.shape}' in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        nifti.apply_atlas(d, a, b)
    assert f'atlas with shape {a.shape}' in str(errorinfo.value)
    assert f'shape {d.shape}' in str(errorinfo.value)


def test_break_unfold_atlas():
    d = rand(6, 5)
    a = rand(2, 3, 4)
    m = rand(3)
    with raises(ValueError) as errorinfo:
        nifti.unfold_atlas(d, a, m)
    assert f'{m.ndim}D mask on {a.ndim}D atlas' in str(errorinfo.value)

    m = rand(3, 3)
    with raises(ValueError) as errorinfo:
        nifti.unfold_atlas(d, a, m)
    assert f'atlas with shape {a.shape}' in str(errorinfo.value)
    assert 'mask with shape (3, 3, 1)' in str(errorinfo.value)

    a = asarray([0, 1, 2])
    with raises(ValueError) as errorinfo:
        nifti.unfold_atlas(d, a)
    assert f'data with shape {d.shape}' in str(errorinfo.value)
    assert '2 parcels' in str(errorinfo.value)



#!/usr/bin/env python3
"""
Operations for timeseries.

Attributes
----------
LGR
    Logger
"""

import logging

import numpy as np

from nigsp.utils import prepare_ndim_iteration

LGR = logging.getLogger(__name__)


def normalise_ts(timeseries):
    """
    Normalise given timeseries (i.e. mean=0, std=1).

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Any timeseries with std == 0 is returned as a series of 0s.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.

    Returns
    -------
    numpy.ndarray
        The normalised timeseries (mean=0 std=1) if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.
    """
    if timeseries.ndim < 2 or (timeseries.ndim == 2 and timeseries.shape[1] == 1):
        LGR.warning('Given timeseries seems to be a single timepoint. '
                    'Returning it as is.')
        return timeseries

    z = ((timeseries - timeseries.mean(axis=1)[:, np.newaxis, ...]) /
         timeseries.std(axis=1, ddof=1)[:, np.newaxis, ...])
    z[np.isnan(z)] = 0

    return z


def graph_fourier_transform(timeseries, eigenvec, energy=False, mean=False):
    """
    Projet a graph decomposition onto the timeseries.

    It returns the result of the projection or the energy of the spectral
    density of the projection.

    If `mean` is true and the timeseries has 3 dimensions, it returns the
    mean across the last dimension.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    eigenvec : numpy.ndarray
        The eigenvector resulting from the Laplacian decomposition.
    energy : bool, optional
        If True, returns the energy (power) of the spectral density instead of
        the projection.
    mean : bool, optional
        If True and timeseries has 3 dimensions, returns the mean across the last axis.

    Returns
    -------
    np.ndarray
        Returns either the projection of the graph on the timeseries, or its energy.
    """
    timeseries = timeseries.squeeze()
    if timeseries.ndim < 3:
        proj = eigenvec.conj().T @ timeseries
    else:
        temp_ts, proj = prepare_ndim_iteration(timeseries, 2)
        for i in range(temp_ts.shape[-1]):
            proj[:, :, i] = eigenvec.conj().T @ np.squeeze(temp_ts[:, :, i])
        if timeseries.ndim > 3:
            proj = proj.reshape(timeseries.shape)

    if timeseries.ndim > 2 and mean and not energy:
        proj = proj.mean(axis=1)

    if energy:
        # Compute energy of the spectral density
        energy = proj ** 2
        if proj.ndim > 2 and mean:
            energy = energy.mean(axis=1)

        return energy

    else:
        return proj


def median_cutoff_frequency_idx(energy):
    """
    Find the frequency that splits the energy of a timeseries in two roughly equal parts.

    Parameters
    ----------
    energy : numpy.ndarray
        The array representing the energy (power) spectral density of a timeseries.
        this array can be 1D or 2D - if 2D it's assumed that the second dimension
        represents subjects.

    Returns
    -------
    int
        The index of the frequency that splits the spectral power into two
        (more or less) equal parts.

    Raises
    ------
    NotImplementedError
        If the provided array is 3D or more.
    """
    if energy.ndim > 2:
        raise NotImplementedError('Provided energy spectral density data have '
                                  f'{energy.ndim} dimensions, but arrays of more '
                                  'than 2 dimensions are not supported yet')

    if energy.ndim == 2:
        energy = energy.mean(axis=-1)
    half_tot_auc = np.trapz(energy, axis=0)/2
    LGR.debug(f'Total AUC = {half_tot_auc*2}, targetting half of total AUC')

    # Compute the AUC from first to one to last frequency,
    # skipping first component because AUC(1)=0.
    # Returns first idx for which AUC reaches half of total AUC.
    # It could be computed from high to low, but in theory
    # there would be more cycles, as #eigenvec(h) > #eigenvec(l).
    for freq_idx in range(1, energy.size):
        LGR.debug(f'Frequency idx {freq_idx}, '
                  f'AUC = {np.trapz(energy[:freq_idx])}, '
                  f'target AUC = {half_tot_auc}')
        if np.trapz(energy[:freq_idx]) >= half_tot_auc:
            break

    LGR.info(f'Found {freq_idx} as splitting index')
    return freq_idx


def graph_filter(timeseries, eigenvec, freq_idx, keys=['low', 'high']):
    """
    Filter a graph decomposition into two parts based on freq_idx.

    Return the two eigenvector lists (high freq and low freq) that are equal
    to the original eigenvector list, but "low" is zero-ed for all frequencies
    >= of the given index, and "high" is zero-ed for all frequencies < to the
    given index.
    Also return their projection onto a timeseries.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    eigenvec : numpy.ndarray
        The eigenvector resulting from the Laplacian decomposition.
    freq_idx : int
        The index of the frequency that splits the spectral power into two
        (more or less) equal parts - i.e. the index of the first frequency in
        the "high" component.
    keys : list, optional
        The keys to call the splitted parts with

    Returns
    -------
    dict of numpy.ndarray
        Return first the split eigenvectors
    dict of numpy.ndarray
        Return second the projected split eigenvectors onto the timeseries.

    Raises
    ------
    IndexError
        If the given index is 0 (all "high"), the last possible index (all "low"),
        or higher than the last possible index (not applicable).
    """
    # #!# Find better name
    # #!# Implement a multi-splitter and an index splitter
    if freq_idx == 0 or freq_idx >= eigenvec.shape[0]-1:
        raise IndexError(f'Selected index {freq_idx} is not valid to '
                         f'split eigenvector matrix of shape {eigenvec.shape}.')

    evec_split = dict.fromkeys(keys)
    ts_split = dict.fromkeys(keys)
    evec_split['low'] = np.append(eigenvec[:, :freq_idx],
                                  np.zeros_like(eigenvec[:, freq_idx:],
                                                dtype='float32'), axis=-1)
    evec_split['high'] = np.append(np.zeros_like(eigenvec[:, :freq_idx],
                                                 dtype='float32'),
                                   eigenvec[:, freq_idx:], axis=-1)

    LGR.info('Compute graph fourier coefficients.')
    fourier_coeff = graph_fourier_transform(timeseries, eigenvec)

    for k in keys:
        LGR.info(f'Compute {k} portion of timeseries.')
        ts_split[k] = graph_fourier_transform(fourier_coeff, evec_split[k].T)

    return evec_split, ts_split


def functional_connectivity(timeseries, mean=False):
    """
    Compute Functional Connectivity of timeseries.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray or dict of numpy.ndarray
        Adictionary of (or a single) 2- or 3-D matrix with timeseries along axis 1.
    mean : bool, optional
        If timeseries is 3D and this is True, return the average FC along the last axis.

    Returns
    -------
    numpy.ndarray or dict of numpy.ndarray
        Functional Connectivity of the given timeseries input.
    """

    def _fc(timeseries, mean=False):
        """
        Quick functional connectivity computation.

        This is not meant to be used outside of other functions.

        It is assumed that time is encoded in the second dimension (axis 1),
        e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

        Parameters
        ----------
        timeseries : numpy.ndarray
            A 2- or 3-D matrix containing timeseries along axis 1.
        mean : bool, optional
            If timeseries is 3D and this is True, return the average FC along the last axis.

        Returns
        -------
        numpy.ndarray
            FC matrix
        """
        timeseries = timeseries.squeeze()
        if timeseries.ndim < 2:
            LGR.warning('Computing functional connectivity of a 1D array (== 1)!')
            return 1
        elif timeseries.ndim == 2:
            return np.corrcoef(timeseries)
        else:
            # reshape the array to allow reiteration on unknown dimensions of timeseries 
            temp_ts, _ = prepare_ndim_iteration(timeseries, 2)
            fcorr = np.empty(([temp_ts.shape[0]] * 2 + [temp_ts.shape[-1]]), dtype='float32')
            for i in range(temp_ts.shape[-1]):
                fcorr[:, :, i] = np.corrcoef(temp_ts[..., i])

            if timeseries.ndim > 3:
                fcorr = fcorr.reshape(timeseries.shape)

            if mean:
                fcorr = fcorr.mean(axis=2)
            return fcorr

    if type(timeseries) is dict:
        fc = dict()
        for k in timeseries.keys():
            fc[k] = _fc(timeseries[k], mean=False)
    else:
        fc = _fc(timeseries, mean=False)

    return fc


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
